/**
Aufbau

- author: Albert Wiedemann
- version: 1.0
*/
  
import UIKit
import PlaygroundSupport

/** Hellblau */
let hellBlau = CGColor(red: 0.9, green: 0.9, blue: 1.0, alpha: 1.0)
/** Hellgrün */
let hellGrün = CGColor(red: 0.9, green: 1.0, blue: 0.9, alpha: 1.0)
/** Hellrot */
let hellRot = CGColor(red: 1.0, green: 0.9, blue: 0.9, alpha: 1.0)
/** Hellgelb */
let hellGelb = CGColor(red: 1.0, green: 1.0, blue: 0.9, alpha: 1.0)
/** Hellgrau */
let hellGrau = CGColor(red: 0.8, green: 0.8, blue: 0.8, alpha: 1.0)
/** Grau */
let grau = CGColor(red: 0.5, green: 0.5, blue: 0.5, alpha: 1.0)
/** Schwarz */
let schwarz = CGColor(red: 0.0, green: 0.0, blue: 0.0, alpha: 1.0)
/** Weiß */
let weiß = CGColor(red: 1.0, green: 1.0, blue: 1.0, alpha: 1.0)
/** Rot */
let rot = CGColor(red: 1.0, green: 0.0, blue: 0.0, alpha: 1.0)
/** Grün */
let grün = CGColor(red: 0.0, green: 1.0, blue: 0.0, alpha: 1.0)
/** Blau */
let blau = CGColor(red: 0.0, green: 0.0, blue: 1.0, alpha: 1.0)
/** Gelb */
let gelb = CGColor(red: 1.0, green: 1.0, blue: 0.0, alpha: 1.0)
/** Cyan */
let cyan = CGColor(red: 0.0, green: 1.0, blue: 1.0, alpha: 1.0)
/** Magenta*/
let magenta = CGColor(red: 1.0, green: 0.0, blue: 1.0, alpha: 1.0)
/** Orange */
let orange = CGColor(red: 255,green: 128,blue: 0, alpha: 1.0)
/** Braun */
let braun = CGColor(red: 128,green: 64,blue: 0, alpha: 1.0)
/** Font für Textausgaben */
let font = CTFontCreateWithName (("Helvetica" as CFString?)!, 10.0, nil)

/**
Beschreibt die verfügbaren Werkzeuge
*/
enum Werkzeuge: CaseIterable
{
    case verschieben, knotenErstellen, kanteErstellen
}

/**
Beschreibt einen Beobachter der Werkzeugleiste
 */
protocol WerkzeugBeobachter
{
    /**
    Meldet, dass sich das aktuelle Werkzeug geändert hat.
    - parameters:
        - aktWerkzeug: das neue, aktuelle Werkzeug
     */
    func AktuellesWerkzeugsetzen(aktWerkzeug: Werkzeuge)
}

/**
Stellt die Werkzeugleiste dar
*/
class Werkzeugleiste: UIView
{
    /** Das aktuell ausgewählte Werkzeug */
    private var aktWerkzeug: Werkzeuge
    /** Die Seitenlänge eines Werkzeugicons */
    private var größe: CGFloat
    /** Symbol bei MouseDown */
    private var downWerkzeug: Werkzeuge
    /** Der Beobachter für Veränderungen des aktuellen Werkzeugs */
    private var beobachter: WerkzeugBeobachter?
    /** Anzahl der Kommandoknöpfe */
    private final let anzahlKnöpfe = 7
    /** Kommandoknopf */
    private var knopfNummer: Int
    /** Oberflächenverwaltung */
    private var oberfläche: OberflächenController?

    /**
    Standardkonstruktor, legt die Größe und Position fest
    - parameters:
        - frame: Framerechteck
    */
    override init(frame: CGRect)
    {
        knopfNummer = -1
        beobachter = nil
        größe = 50
        aktWerkzeug = .verschieben
        downWerkzeug = .verschieben
        oberfläche = nil
        super.init(frame: frame)
        größe = bounds.height
    }
    
    /**
    Standardkonstruktor, liest daten über den Decoder
    - parameters:
        - coder: Dekodierobjekt
    */
    required init?(coder: NSCoder)
    {
        knopfNummer = -1
        beobachter = nil
        größe = 50
        aktWerkzeug = .verschieben
        downWerkzeug = .verschieben
        super.init(coder: coder)
        größe = bounds.height
    }
    
    /**
    Aufgerufen, wenn ein oder mehr Finger die Tastatur berühren.
    - parameters:
        - touches: Menge an Berüchrungsbeschreibungen für den Beginn des Ereignisses
        - event: das Gesamtereignis, zu dem die Berührungen gehören
    */
    override func touchesBegan(_ touches: Set<UITouch>, with event: UIEvent?) {
        for t in touches as Set<UITouch>
        {
            let downPunkt = t.location(in: self)
            if downPunkt.x < CGFloat(Werkzeuge.allCases.count) * größe
            {
                downWerkzeug = Werkzeuge.allCases[Int(downPunkt.x / größe)]
                knopfNummer = -1
            }
            else
            {
                downWerkzeug = aktWerkzeug
                if downPunkt.x >= bounds.width - CGFloat(anzahlKnöpfe) * größe
                {
                    knopfNummer = Int((downPunkt.x - (bounds.width - CGFloat(anzahlKnöpfe) * größe)) / größe)
                }
                else
                {
                    knopfNummer = -1
                }
                setNeedsDisplay()
            }
        }
        super.touchesBegan(touches, with: event)
    }

    /**
    Aufgerufen, wenn ein oder mehr Finger die Tastatur verlassen.
    - parameters:
        - touches: Menge an Berührungsbeschreibungen für den Beginn des Ereignisses
        - event: das Gesamtereignis, zu dem die Berührungen gehören
    */
    override func touchesEnded(_ touches: Set<UITouch>, with event: UIEvent?)
    {
        for t in touches as Set<UITouch>
        {
            let downPunkt = t.location(in: self)
            if (downPunkt.x >= 0) && (downPunkt.x < CGFloat(Werkzeuge.allCases.count) * größe) && (downPunkt.y >= 0) && (downPunkt.y < größe)
            {
                if downWerkzeug == Werkzeuge.allCases[Int(downPunkt.x / größe)]
                {
                    aktWerkzeug = downWerkzeug
                    if beobachter != nil
                    {
                        beobachter!.AktuellesWerkzeugsetzen(aktWerkzeug: aktWerkzeug)
                    }
                }
            }
            else if (downPunkt.x >= bounds.width - CGFloat(anzahlKnöpfe) * größe) && (downPunkt.x < bounds.width) && (downPunkt.y >= 0) && (downPunkt.y < größe)
            {
                if Int((downPunkt.x - (bounds.width - CGFloat(anzahlKnöpfe) * größe)) / größe) == knopfNummer
                {
                    switch knopfNummer
                    {
                        case 0:
                            if beobachter != nil
                            {
                                oberfläche!.Löschen()
                            }
                        case 1:
                        	oberfläche!.EditierenAnzeigen()
                        case 2:
                        	oberfläche!.ZurücksetzenAusführen()
                        case 3:
                        	oberfläche!.LesenAusführen()
                        case 4:
                        	oberfläche!.SichernAusführen(immer: false)
                        case 5:
                        	oberfläche!.AusgebenAusführen()
                        case 6:
                        	oberfläche!.HilfeAusführen()
                        default:
                            break
                    }
                }
            }
            knopfNummer = -1
            setNeedsDisplay()
        }
        super.touchesEnded(touches, with: event)
    }

    /**
    Sollte eigentlich gar nicht aufgerufen werden.
    */
    override func tintColorDidChange()
    {
        setNeedsDisplay()
    }

    /**
    Die Zeichenmethode wird immer dann aufgrufen, wenn edie View oder ein Teil davon neu gezeichnet werden muss.
    - parameters:
        - rect: das (mindestens) neu zu zeichnende Rechteck
    */
    override func draw(_ rect: CGRect)
    {
        /**
        Zeichnet den Text
        - parameters:
            - text: der zu zeichnende Text
            - context: der Grafikcontext
            - pos: die x-Position des Textes
         */
        func TextZeichnen (text: String, context: CGContext, pos: CGFloat)
        {
            let attrString = NSAttributedString(string: text, attributes: [(NSAttributedString.Key(rawValue: kCTForegroundColorAttributeName as String as String)):UIColor.black, (NSAttributedString.Key(rawValue: kCTFontAttributeName as String as String)):font])
            let line = CTLineCreateWithAttributedString(attrString)
            let bound = CTLineGetImageBounds(line, context)
            context.textPosition = CGPoint(x: pos + größe / 2 - bound.width / 2, y: größe / 2 - 2)
            CTLineDraw(line, context)

        }
    
        if let context = UIGraphicsGetCurrentContext()
        {
            context.setFillColor(CGColor(red: 0.9, green: 0.9, blue: 0.9, alpha: 1.0))
            context.fill(rect)
            var pos = 0.0
            let delta = größe / 10
            context.textMatrix  = CGAffineTransform (a: 1.0, b: 0.0, c: 0.0, d: -1.0, tx: 0.0, ty: 0.0)
            
            for w in Werkzeuge.allCases
            {
                context.setFillColor(w == aktWerkzeug ? hellGrün : hellGrau)
                var rect = CGRect(x: pos, y: 0.0, width: größe, height: größe)
                context.fill(rect)
                context.setStrokeColor(schwarz)
                context.stroke(rect)
                context.setFillColor(weiß)
                context.fill(rect.insetBy(dx: delta, dy: delta))
                context.stroke(rect.insetBy(dx: delta, dy: delta))
                context.setFillColor(grau)
                
                let p = CGPath(rect: CGRect(x: pos + größe / 2 - 2 * delta, y: 2 * delta, width: 4 * delta, height: größe - 4 * delta), transform: nil)
                switch w
                {
                    case .verschieben:
                        context.move(to: CGPoint(x: pos + größe / 4, y:  größe * 3 / 4))
                        context.addLine(to: CGPoint(x: pos + größe * 3 / 4, y: größe / 4))
                        context.move(to: CGPoint(x: pos + größe * 5 / 8, y:  größe / 4))
                        context.addLine(to: CGPoint(x: pos + größe * 3 / 4, y: größe / 4))
                        context.move(to: CGPoint(x: pos + größe * 3 / 4, y:  größe * 3 / 8))
                        context.addLine(to: CGPoint(x: pos + größe * 3 / 4, y: größe / 4))
                        context.strokePath()
					case .knotenErstellen:
                		context.setFillColor(hellRot)
                    	context.fillEllipse(in: CGRect(x: pos + größe / 4, y: größe / 4, width: größe / 2, height: größe / 2))
						context.addPath(CGPath(ellipseIn: CGRect(x: pos + größe / 4, y: größe / 4, width: größe / 2, height: größe / 2), transform: nil))
                    	context.setStrokeColor(rot)
                    	context.strokePath()
                    	context.setStrokeColor(schwarz)
                    case .kanteErstellen:
                    	let r = größe / 10
                        context.addPath(CGPath(ellipseIn: CGRect(x: pos + delta + r, y: größe / 2 - r, width: r * 2, height: r * 2), transform: nil))
                        context.addPath(CGPath(ellipseIn: CGRect(x: pos + größe - delta - 3 * r, y: größe / 2 - r, width: r * 2, height: r * 2), transform: nil))
                        context.strokePath()
                        context.move(to: CGPoint(x: pos + delta + 3 * r, y: größe / 2))
                        context.addLine(to: CGPoint(x: pos + größe - delta - 3 * r, y: größe / 2))
                    	context.setStrokeColor(rot)
                    	context.strokePath()
                    	context.setStrokeColor(schwarz)
                }
                pos += größe;
            }
            pos = self.bounds.width - größe * CGFloat(anzahlKnöpfe)
            for i in 0 ..< anzahlKnöpfe
            {
                context.setFillColor(i == knopfNummer ? weiß : hellGrau)
                var rect = CGRect(x: pos, y: 0.0, width: größe, height: größe)
                context.fill(rect)
                context.setStrokeColor(schwarz)
                context.stroke(rect)

                switch (i)
                {
                    case 0:
                        context.setFillColor(hellRot)
                        context.addPath(CGPath(roundedRect: CGRect(x: pos + 12, y: 10, width: größe - 24, height: 5), cornerWidth: 4, cornerHeight: 4, transform: nil))
                        context.fillPath()
                        context.fill(CGRect(x: pos + 16, y: 15, width: größe - 32, height: größe - 24))
                        context.stroke(CGRect(x: pos + 16, y: 15, width: größe - 32, height: größe - 24))
                        context.addPath(CGPath(roundedRect: CGRect(x: pos + 12, y: 10, width: größe - 24, height: 5), cornerWidth: 4, cornerHeight: 4, transform: nil))
                        context.move(to: CGPoint(x: pos + größe / 2, y: 15))
                        context.addLine(to: CGPoint(x: pos + größe / 2, y: größe - 10))
                        context.move(to: CGPoint(x: pos + größe / 4 + 8, y: 15))
                        context.addLine(to: CGPoint(x: pos + größe / 4 + 8, y: größe - 10))
                        context.move(to: CGPoint(x: pos +  3 * größe / 4 - 8, y: 15))
                        context.addLine(to: CGPoint(x: pos + 3 * größe / 4 - 8, y: größe - 10))
                        context.strokePath()
                    case 1:
                        context.setFillColor(hellRot)
                        context.fill(CGRect(x: pos + größe / 10, y: größe / 10, width: größe * 8 / 10, height: größe * 8 / 10))
                        context.stroke(CGRect(x: pos + größe / 10, y: größe / 10, width: größe * 8 / 10, height: größe * 8 / 10))
                        context.setFillColor(weiß)
                        context.fill(CGRect(x: pos + größe / 5, y: größe / 5, width: größe * 3 / 5, height: größe * 3 / 5))
                        context.setFillColor(grau)
                        context.fill(CGRect(x: pos + größe / 5 + 3, y: größe / 5 + 3, width: größe * 3 / 5 - 6, height: 3))
                        context.fill(CGRect(x: pos + größe / 5 + 3, y: größe / 5 + 9, width: größe * 3 / 5 - 6, height: 3))
                        context.fill(CGRect(x: pos + größe / 5 + 3, y: größe / 5 + 15, width: größe * 3 / 5 - 6, height: 3))
                        context.setFillColor(blau)
                        context.fill(CGRect(x: pos + größe / 2 - 10, y: größe / 5 + 21, width: 6, height: 3))
                        context.fill(CGRect(x: pos + größe / 2 + 4, y: größe / 5 + 21, width: 6, height: 3))
                    case 2:
                        context.setFillColor(hellGelb)
                        context.move(to: CGPoint(x: pos + größe / 5, y: größe / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 7 / 10, y: größe / 10))
                         context.addLine(to: CGPoint(x: pos + größe * 4 / 5, y: größe / 5))
                        context.addLine(to: CGPoint(x: pos + größe * 4 / 5, y: größe * 9 / 10))
                        context.addLine(to: CGPoint(x: pos + größe / 5, y: größe * 9 / 10))
                        context.addLine(to: CGPoint(x: pos + größe / 5, y: größe / 10))
                        let p = context.path
                        context.fillPath()
                        context.addPath(p!)
                        context.strokePath()
                    case 3:
                        context.setFillColor(hellGelb)
                        context.move(to: CGPoint(x: pos + größe * 2 / 10, y: größe * 2 / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 5 / 20, y: größe * 3 / 20))
                        context.addLine(to: CGPoint(x: pos + größe * 17 / 20, y: größe * 7 / 20))
                        context.addLine(to: CGPoint(x: pos + größe * 17 / 20, y: größe * 15 / 20))
                        context.addLine(to: CGPoint(x: pos + größe * 8 / 10, y: größe * 8 / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 2 / 10, y: größe * 6 / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 2 / 10, y: größe * 2 / 10))
                        context.fillPath()
                        context.move(to: CGPoint(x: pos + größe * 2 / 10, y: größe * 2 / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 8 / 10, y: größe * 4 / 10))
                        context.move(to: CGPoint(x: pos + größe * 8 / 10, y: größe * 4 / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 8 / 10, y: größe * 8 / 10))
                        context.move(to: CGPoint(x: pos + größe * 8 / 10, y: größe * 8 / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 2 / 10, y: größe * 6 / 10))
                        context.move(to: CGPoint(x: pos + größe * 2 / 10, y: größe * 6 / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 2 / 10, y: größe * 2 / 10))

                        context.move(to: CGPoint(x: pos + größe * 2 / 10, y: größe * 2 / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 5 / 20, y: größe * 3 / 20))
                        context.move(to: CGPoint(x: pos + größe * 5 / 20, y: größe * 3 / 20))
                        context.addLine(to: CGPoint(x: pos + größe * 17 / 20, y: größe * 7 / 20))
                        context.move(to: CGPoint(x: pos + größe * 8 / 10, y: größe * 4 / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 17 / 20, y: größe * 7 / 20))
                        context.move(to: CGPoint(x: pos + größe * 17 / 20, y: größe * 7 / 20))
                        context.addLine(to: CGPoint(x: pos + größe * 17 / 20, y: größe * 15 / 20))
                        context.move(to: CGPoint(x: pos + größe * 8 / 10, y: größe * 8 / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 17 / 20, y: größe * 15 / 20))
                        context.strokePath()
                    case 4:
                        context.setFillColor(hellGrün)
                        context.move(to: CGPoint(x: pos + größe * 2 / 10, y: größe * 2 / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 7 / 10, y: größe * 2 / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 8 / 10, y: größe * 3 / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 8 / 10, y: größe * 8 / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 2 / 10, y: größe * 8 / 10))
                        context.fillPath()
                        
                        context.setFillColor(weiß)
                        context.fill(CGRect(x: pos + größe * 4 / 10, y: größe * 2 / 10, width: größe * 2 / 10, height: größe * 1 / 10))

                        context.move(to: CGPoint(x: pos + größe / 5, y: größe * 2 / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 7 / 10, y: größe * 2 / 10))
                        context.move(to: CGPoint(x: pos + größe * 7 / 10, y: größe * 2 / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 4 / 5, y: größe * 3 / 10))
                        context.move(to: CGPoint(x: pos + größe * 4 / 5, y: größe * 3 / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 4 / 5, y: größe * 8 / 10))
                        context.move(to: CGPoint(x: pos + größe * 4 / 5, y: größe * 8 / 10))
                        context.addLine(to: CGPoint(x: pos + größe / 5, y: größe * 8 / 10))
                        context.move(to: CGPoint(x: pos + größe / 5, y: größe * 8 / 10))
                        context.addLine(to: CGPoint(x: pos + größe / 5, y: größe * 2 / 10))
                        context.addEllipse(in: CGRect(x: pos + größe * 4 / 10, y: größe * 4 / 10, width: größe * 2 / 10, height: größe * 2 / 10))
                        context.strokePath()
                        context.addEllipse(in: CGRect(x: pos + größe * 9 / 20, y: größe * 9 / 20, width: größe * 2 / 20, height: größe * 2 / 20))
                        context.setFillColor(schwarz)
                        context.fillPath()
                        context.move(to: CGPoint(x: pos + größe * 4 / 10, y: größe * 2 / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 4 / 10, y: größe * 3 / 10))
                        context.move(to: CGPoint(x: pos + größe * 4 / 10, y: größe * 3 / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 6 / 10, y: größe * 3 / 10))
                        context.move(to: CGPoint(x: pos + größe * 6 / 10, y: größe * 3 / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 6 / 10, y: größe * 2 / 10))
                        context.strokePath()
                    case 5:
                        context.setFillColor(hellBlau)
                        context.addPath(CGPath(roundedRect: CGRect(x: pos + größe / 5, y: größe * 2 / 10, width: größe / 2, height: größe * 6 / 10), cornerWidth: 4, cornerHeight: 4, transform: nil))
                        context.fillPath()
                        context.addPath(CGPath(roundedRect: CGRect(x: pos + größe / 5, y: größe * 2 / 10, width: größe / 2, height: größe * 6 / 10), cornerWidth: 4, cornerHeight: 4, transform: nil))
                        context.strokePath()
                        context.setStrokeColor(hellBlau)
                        context.move(to: CGPoint(x: pos + größe * 7 / 10, y: größe * 4 / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 7 / 10, y: größe * 6 / 10))
                        context.strokePath()
                        context.setStrokeColor(schwarz)
                        context.move(to: CGPoint(x: pos + größe * 5 / 10, y: größe / 2))
                        context.addLine(to: CGPoint(x: pos + größe * 9 / 10, y: größe / 2))
                        context.move(to: CGPoint(x: pos + größe * 8 / 10, y: größe * 4 / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 9 / 10, y: größe / 2))
                        context.move(to: CGPoint(x: pos + größe * 8 / 10, y: größe * 6 / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 9 / 10, y: größe / 2))
                        context.strokePath()
                    case 6:
                        context.move(to: CGPoint(x: pos + größe / 5, y: größe / 10))
                        context.addLine(to: CGPoint(x: pos + größe / 5, y: größe * 9 / 10))
                        context.move(to: CGPoint(x: pos + größe / 5 + 1, y: größe / 10))
                        context.addLine(to: CGPoint(x: pos + größe / 5 + 1, y: größe * 9 / 10))
                        context.move(to: CGPoint(x: pos + größe / 5, y: größe / 2))
                        context.addLine(to: CGPoint(x: pos + größe * 4 / 5, y: größe / 2))
                        context.move(to: CGPoint(x: pos + größe / 5, y: größe / 2 - 1))
                        context.addLine(to: CGPoint(x: pos + größe * 4 / 5, y: größe / 2 - 1))
                        context.move(to: CGPoint(x: pos + größe * 4 / 5, y: größe / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 4 / 5, y: größe * 9 / 10))
                        context.move(to: CGPoint(x: pos + größe * 4 / 5 - 1, y: größe / 10))
                        context.addLine(to: CGPoint(x: pos + größe * 4 / 5 - 1, y: größe * 9 / 10))
                        context.strokePath()
                    default:
                        break
                }
                pos += größe;
            }
        }
    }
    
    /**
    Trägt den Beobachter für die Werkzeugänderungen ein.
    - parameters:
        - b: der Beobachter
    */
    func BeobachterAnmelden(b: WerkzeugBeobachter)
    {
        beobachter = b
    }
    
    /**
    Trägt die Oberfläche ein.
    - parameters:
        - o: die Oberflächenverwaltung
    */
    func OberflächeSetzen(o: OberflächenController)
    {
        oberfläche = o
    }
}

/**
Die Zeichenfläche zur Darstellung des Graphen
*/
class GraphenAnzeige: UIView
{
    /** Feld aller zu zeichnenden Objekte. */
    private var alleSymbole: [GraphenSymbol]
    /** Zeitstempel für Mausevents */
    private var klickZeit: TimeInterval?
    /** Empfänger für Oberflächenereignisse */
    private var empfänger: EreignisEmpfänger?

    /**
    Standardkonstruktor, legt die Größe und Position fest
    - parameters:
        - frame: Framerechteck
    */
    override init(frame: CGRect)
    {
        alleSymbole = [GraphenSymbol]()
        klickZeit = 0
        empfänger = nil
        super.init(frame: frame)
    }
    
    /**
    Standardkonstruktor, liest daten über den Decoder
    - parameters:
        - coder: Dekodierobjekt
    */
    required init?(coder: NSCoder)
    {
        alleSymbole = [GraphenSymbol]()
        klickZeit = 0
        empfänger = nil
        super.init(coder: coder)
    }

    /**
    Zeichnet die Schaltung
    - parameters:
        - rect: das eventuell eingeschränkte Rechteck
    */
    override func draw(_ rect: CGRect)
    {
        if let context = UIGraphicsGetCurrentContext()
        {
			context.textMatrix  = CGAffineTransform (a: 1.0, b: 0.0, c: 0.0, d: -1.0, tx: 0.0, ty: 0.0)
            context.setFillColor(hellGrau)
        	context.fill(CGRect(x: 0, y: 0, width: bounds.width, height: bounds.height))
        	context.setStrokeColor(schwarz)
            let a = alleSymbole
            for s in a
            {
            	s.Zeichnen(context: context)
            }
        }
    }

    /**
    Trägt den Empfänger für die Ereignisse ein.
    - parameters:
        - e: der Ereignisempfänger
     */
    func EmpfängerAnmelden(e: EreignisEmpfänger)
    {
        empfänger = e
    }
    
    /**
    Aufgerufen, wenn ein oder mehr Finger die Tastatur berühren.
    - parameters:
        - touches: Menge an Berüchrungsbeschreibungen für den Beginn des Ereignisses
        - event: das Gesamtereignis, zu dem die Berührungen gehören
    */
    override func touchesBegan(_ touches: Set<UITouch>, with event: UIEvent?) {
        klickZeit = event?.timestamp
        for t in touches as Set<UITouch>
        {
            let downPunkt = t.location(in: self)
            empfänger?.TouchDown(x: downPunkt.x, y: downPunkt.y)
        }
        super.touchesBegan(touches, with: event)
    }
    
    /**
    Aufgerufen, wenn ein gedrückter Finger bewegt wird.
    - parameters:
        - touches: Menge an Berüchrungsbeschreibungen für den Beginn des Ereignisses
        - event: das Gesamtereignis, zu dem die Berührungen gehören
    */
    override func touchesMoved(_ touches: Set<UITouch>, with event: UIEvent?) {
        let zeit = event?.timestamp
        if (zeit == nil) || (klickZeit == nil) || (zeit! - klickZeit! > 0.2)
        {
            for t in touches as Set<UITouch>
            {
                let downPunkt = t.location(in: self)
                empfänger?.TouchGezogen(x: downPunkt.x, y: downPunkt.y)
            }
        }
        super.touchesMoved(touches, with: event)
    }

    /**
    Aufgerufen, wenn ein oder mehr Finger die Tastatur verlassen.
    - parameters:
        - touches: Menge an Berührungsbeschreibungen für den Beginn des Ereignisses
        - event: das Gesamtereignis, zu dem die Berührungen gehören
    */
    override func touchesEnded(_ touches: Set<UITouch>, with event: UIEvent?)
    {
        for t in touches as Set<UITouch>
        {
            let downPunkt = t.location(in: self)
            empfänger?.TouchUp(x: downPunkt.x, y: downPunkt.y)
        }
        super.touchesEnded(touches, with: event)
    }

    /**
	Trägt das angegebene Symbol in die Liste ein.
	- parameters:
		- symbol: das einzutragende Symbol
     */
	func Eintragen(symbol: GraphenSymbol)
	{
		alleSymbole.append(symbol)
    	setNeedsDisplay()
    }
    
    /**
	Entfernt das angegebene Symbol.
	- parameters:
		- symbol: das zu entfernende Symbol
    */
    func Entfernen(symbol: GraphenSymbol)
    {
        alleSymbole.removeAll(where: {$0 === symbol})
        setNeedsDisplay()
    }

	/**
	Meldet die Symbole zurück
	- returns: das Feld der Symbole
    */
    func SymboleGeben() -> [GraphenSymbol]
    {
    	return alleSymbole;
    }
}

/**
Oberklasse der Symbole des Graphen
*/
class GraphenSymbol
{

    /** Die Oberfläche, die das Symbol verwaltet */
    fileprivate let anzeige: GraphenAnzeige
    
    /**
	Setzt die zugeordnete View
	- parameters:
		- a: die zugeordnete View
    */
    init(a: GraphenAnzeige)
    {
        anzeige = a
        a.Eintragen(symbol: self)
    }
    
    /**
    Zeichnet den Text
    - parameters:
        - text: der zu zeichnende Text
        - context: der Grafikcontext
        - xpos: die x-Position des Textes
        - ypos: die y-Position des Textes
    */
    func TextZeichnen(text: String, context: CGContext, xpos: CGFloat, ypos: CGFloat)
    {
        let attrString = NSAttributedString(string: text, attributes: [(NSAttributedString.Key(rawValue: kCTForegroundColorAttributeName as String as String)):UIColor.black, (NSAttributedString.Key(rawValue: kCTFontAttributeName as String as String)):font])
        let line = CTLineCreateWithAttributedString(attrString)
        let bound = CTLineGetImageBounds(line, context)
        context.textPosition = CGPoint(x: xpos - bound.width / 2, y: ypos - 2)
        CTLineDraw(line, context)
    }

    /**
	Stellt das Symbol dar.
	- parameters:
		- context: das Graphikobjekt zum Zeichnen
     */
    func Zeichnen(context: CGContext)
    {
    	//hier nichts zu tun
    }
    
    /**
	Bestimmt die RGB-Farbe für den gegeben String.
	- parameters:
		- farbe: die Farbe als String
	- returns: die Farbe als RGB-Farbe
    */
    func FarbeCodieren (farbe: String) -> CGColor
    {
        switch (farbe.localizedLowercase)
        {
            case "weiß", "weiss":
                return weiß
            case "rot":
                return rot
            case "grün", "gruen":
                return grün
            case "blau":
                return blau
            case "gelb":
                return gelb
            case "magenta":
                return magenta
            case "cyan":
                return cyan
            case "hellgelb":
                return hellGelb
            case "hellgrün", "hellgruen":
                return hellGrün
            case "orange":
                return orange;
            case "braun":
                return braun;
            case "grau":
                return grau
            case "schwarz":
                return schwarz
            default:
                return schwarz
        }
    }
}

/**
Verwaltet ein Knotensymbol in einem Graph
*/
class KnotenSymbol: GraphenSymbol
{
    /** Der Bezeichner */
    private var bezeichner: String
    /** Radius */
    private var r: CGFloat
    /** x-Koordinate */
    private var x: CGFloat
    /** y-Koordinate */
    private var y: CGFloat
    /** Farbe */
    private var farbe: String
    
    /**
	Legt das Symbol an und besetzt die Attribute.
	- parameters:
		- x: x-Koordinate
		- y: y-Koordinate
		- radius: Radius
		- farbe: Farbe
		- bezeichner: Bezeichner
		- a: die zugeordnete View
    */
    init(x: CGFloat, y: CGFloat, radius: CGFloat, farbe: String, bezeichner: String, a: GraphenAnzeige)
    {
        self.x = x
        self.y = y
        self.r = radius
        self.farbe = farbe
        self.bezeichner = bezeichner
        super.init(a: a)
    }

    /**
	Stellt das Symbol dar.
	- parameters:
		- context das Graphikobjekt zum Zeichnen
    */
    override func Zeichnen(context: CGContext)
    {
        context.setFillColor(FarbeCodieren(farbe: farbe))
        context.fillEllipse(in: CGRect(x: x - r, y: y - r, width: r * 2, height: r * 2))
        context.setStrokeColor(schwarz)
        context.addEllipse(in: CGRect(x: x - r, y: y - r, width: r * 2, height: r * 2))
        context.strokePath()
        TextZeichnen(text: bezeichner, context: context, xpos: x, ypos: y)
    }
    
    /**
	Setzt die Farbe der Darstellung
	- parameters:
		- f: die (neue) Farbe
    */
    func FarbeSetzen(f: String)
    {
        farbe = f
        anzeige.setNeedsDisplay()
    }
    
    /**
	Setzt die Position der Darstellung
	- parameters:
		- x: x-Koordinate
		- y: y-Koordinate
     */
    func PositionSetzen(x: CGFloat, y: CGFloat)
    {
        self.x = x
        self.y = y
        anzeige.setNeedsDisplay()
    }
    
    /**
	Setzt die Farbe der Darstellung
	- parameters:
		- bezeichner: der (neue) Bezeichner
    */
    func BezeichnerSetzen(bezeichner: String)
    {
        self.bezeichner = bezeichner
        anzeige.setNeedsDisplay()
    }
    
    /**
	Meldet die x-Koordinate des Symbols zurück.
	- returns: x-Koordinate
    */
    func XGeben() -> CGFloat
    {
        return x
    }
    
    /**
	Meldet die y-Koordinate des Symbols zurück.
	- returns: y-Koordinate
    */
    func YGeben() -> CGFloat
    {
        return y
    }
    
    /**
	Meldet den Radius des Symbols zurück.
	- returns: Radius
    */
    func RadiusGeben() -> CGFloat
    {
        return r
    }
    
    /**
	Meldet den Bezeichner des Symbols zurück.
	- returns: Bezeichner
    */
    func BezeichnerGeben() -> String
    {
        return bezeichner
    }
}
/**
Verwaltet ein Kantensymbol in einem Graph
*/
class KantenSymbol: GraphenSymbol
{
    /** Breite des Darstellungssymbols */
    private let breite: Int
    /** Gerichtete Kante */
    private var gerichtet: Bool
    /** Gewicht der Kante */
    private var gewicht: String
    /** Farbe */
    private var farbe: String
    /** Startknoten */
    private let von: KnotenSymbol
    /** Endknoten */
    private let nach: KnotenSymbol

    /**
	Legt das Symbol an und besetzt die Attribute.
	- parameters:
		- start: Startknoten
		- ende: Endknoten
		- gerichtet: wenn wahr, ist der Weg gerichtet
		- gewicht: Kantengewicht
		- breite: Breite der Linie
		- farbe: Farbe der Linie
		- a: die zugeordnete View
     */
    init(start: KnotenSymbol, ende: KnotenSymbol, gerichtet: Bool, gewicht: String, breite: Int, farbe: String, a: GraphenAnzeige)
    {
        von = start
        nach = ende
        self.farbe = farbe
        self.breite = breite
        self.gerichtet = gerichtet
        self.gewicht = gewicht
        super.init(a: a)
    }

    /**
	Stellt das Symbol dar.
	- parameters:
		- context: das Graphikobjekt zum Zeichnen
    */
    override func Zeichnen(context: CGContext)
    {
        let x1 = von.XGeben()
        let y1 = von.YGeben()
        let x2 = nach.XGeben()
        let y2 = nach.YGeben()
        let länge = sqrt((x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2))
        var winkel = 0.0
        if x1 == x2
        {
            if y2 > y1
            {
                winkel = -90
            }
            else
            {
                winkel = 90
            }
        }
        else if x1 < x2
        {
            winkel = -atan((y2 - y1) / (x2 - x1)) / Double.pi * 180.0
        }
        else
        {
            winkel = 180.0 - atan((y2 - y1) / (x2 - x1)) / Double.pi * 180.0
        }

        
        var r = von.RadiusGeben();
        var x1neu = x1 + (x2 - x1) / länge * r
        var y1neu = y1 + (y2 - y1) / länge * r
        r = nach.RadiusGeben()
        var x2neu = x2 - (x2 - x1) / länge * r
        var y2neu = y2 - (y2 - y1) / länge * r
        context.setStrokeColor(FarbeCodieren(farbe: farbe))
        context.saveGState()
        context.setLineWidth(CGFloat(breite))
        context.move(to: CGPoint(x: x1neu, y: y1neu))
        context.addLine(to: CGPoint(x: x2neu, y: y2neu))
        if gerichtet
        {
            context.move(to: CGPoint(x: x2neu + 10.0 * cos((-winkel + 135) / 180.0 * Double.pi), y: y2neu + 10.0 * sin((-winkel + 135) / 180.0 * Double.pi)))
            context.addLine(to: CGPoint(x: x2neu, y: y2neu))
            context.move(to: CGPoint(x: x2neu + 10.0 * cos((-winkel - 135) / 180.0 * Double.pi), y: y2neu + 10.0 * sin((-winkel - 135) / 180.0 * Double.pi)))
            context.addLine(to: CGPoint(x: x2neu, y: y2neu))
        }
        context.strokePath()
        context.restoreGState()
        
        let xm = (x1neu + x2neu) / 2
        let ym = (y1neu + y2neu) / 2
        let delta = Double(breite)
        let w = (winkel + 90) / 180.0 * Double.pi
        if (-90 <= winkel) && (winkel <= 0)
        {
            TextZeichnen(text: gewicht, context: context, xpos: xm + delta * cos(w), ypos: ym - delta * sin(w))
        }
        else if (0 < winkel) && (winkel <= 90)
        {
            TextZeichnen(text: gewicht, context: context, xpos: xm - delta * cos(w), ypos: ym + delta * sin(w))
        }
        else if (90 < winkel) && (winkel <= 180)
        {
            TextZeichnen(text: gewicht, context: context, xpos:  xm - delta * cos(w), ypos: ym + delta * sin(w))
        }
        else
        {
            TextZeichnen(text: gewicht, context: context, xpos: xm + delta * cos(w), ypos: ym - delta * sin(w) - 10)
        }
    }
    
    /**
	Setzt die Farbe der Darstellung
	- parameters:
		- f: die (neue) Farbe
    */
    func FarbeSetzen(f: String)
    {
        farbe = f
        anzeige.setNeedsDisplay()
    }
    
    /**
	Setzt das Gewicht der Darstellung
	- parameters:
		- g: das (neue) Gewicht
    */
    func GewichtSetzen(g: String)
    {
        gewicht = g
        anzeige.setNeedsDisplay()
    }
    
    /**
	Setzt, ob die Kante gerichtet ist.
	- parameters:
		- gerichtet: die Kante ist gerichtet bei wahr
    */
    func GerichtetSetzen(gerichtet: Bool)
    {
        self.gerichtet = gerichtet
        anzeige.setNeedsDisplay()
    }
    
    /**
	Meldet den Startknoten
	- returns: Startknoten
    */
    func StartsymbolGeben() -> KnotenSymbol
    {
        return von
    }
    
    /**
	Meldet den Zielknoten
	- returns: Zielknoten
    */
    func ZielsymbolGeben() -> KnotenSymbol
    {
        return nach
    }
    
    /**
	Meldet das Gewicht
	- returns: Gewicht
     */
    func GewichtGeben() -> String
    {
        return gewicht
    }
    
    /**
	Meldet, ob die Kante gerichtet ist
	- returns: wahr, wenn die Kante gerichtet ist
    */
    func IstGerichtet() -> Bool
    {
        return gerichtet
    }
}

/**
Verwaltet ein Symbol für einen gezogenen Knoten in einem Graph
*/
class GezogenerKnotenSymbol: GraphenSymbol
{
    /** Radius */
    private let r: CGFloat
    /** x-Koordinate */
    private var x: CGFloat
    /** y-Koordinate */
    private var y: CGFloat
    
    /**
	Legt das Symbol an und besetzt die Attribute.
	- parameters:
		- x: x-Koordinate
		- y: y-Koordinate
		- radius: Radius
		- a: die zugeordnete View
    */
    init(x: CGFloat, y: CGFloat, radius: CGFloat, a: GraphenAnzeige)
    {
        self.x = x
        self.y = y
        self.r = radius
        super.init(a: a)
    }

    /**
	Stellt das Symbol dar.
	- parameters:
		- context: das Graphikobjekt zum Zeichnen
    */
    override func Zeichnen(context: CGContext)
    {
        context.setFillColor(CGColor(red: 1.0, green: 1.0, blue: 1.0, alpha: 0.4))
        context.fillEllipse(in: CGRect(x: x - r, y: y - r, width: 2 * r, height: 2 * r))
        context.setStrokeColor(CGColor(red: 0.0, green: 0.0, blue: 0.0, alpha: 0.4))
        context.strokeEllipse(in: CGRect(x: x - r, y: y - r, width: 2 * r, height: 2 * r))
    }
    
    /**
	Setzt die Position der Darstellung
	- parameters:
		- x: x-Koordinate
		- y: y-Koordinate
     */
    func PositionSetzen(x: CGFloat, y: CGFloat)
    {
        self.x = x
        self.y = y
        anzeige.setNeedsDisplay()
    }
    
    /**
	Meldet die x-Koordinate des Symbols zurück.
	- returns: x-Koordinate
     */
    func XGeben() -> CGFloat
    {
        return x
    }
    
    /**
	Meldet die y-Koordinate des Symbols zurück.
	- returns: y-Koordinate
    */
    func YGeben() -> CGFloat
    {
        return y
    }
    
    /**
	Meldet den Radius des Symbols zurück.
	- returns: Radius
    */
    func RadiusGeben() -> CGFloat
    {
        return r
    }
}

/**
Beschreibt die Botschaften der Oberfläche an den Empfänger der Ereignisse
 */
protocol EreignisEmpfänger
{
    /**
    Die Oberfläche wurde gedrückt.
    - parameters:
        - x: x-Koordinate des Events
        - y: y-Koordinate des Events
    */
    func TouchDown(x: CGFloat, y: CGFloat)

    /**
    Die Oberfläche wurde losgelassen.
    - parameters:
        - x: x-Koordinate des Events
        - y: y-Koordinate des Events
    */
    func TouchUp(x: CGFloat, y: CGFloat)
    
    /**
    Die Oberfläche wurde gedrückt gezogen.
    - parameters:
        - x: x-Koordinate des Events
        - y: y-Koordinate des Events
    */
    func TouchGezogen(x: CGFloat, y: CGFloat)
    
    /**
    Löscht das aktuell ausgewählte Objekt
     */
    func Löschen()
    
    /**
    Meldet zurück, ob das Szenario verändert wurde.
	- returns: wahr, wenn das Szenario verändert wurde
     */
    func IstGeändert() -> Bool
    
    /**
    Löscht das Szenarium
    */
    func ZurückSetzen()
    
    /**
	Speichert die Datei am angegebenen Ort.

	Eine exisiterende Datei wird gegebenenfalls überschrieben, eine Datenbank zuerst gelöscht.
	- parameters:
		- datei: Dateiname
		- alsDatenbank: gibt an, ob aus einer Datenbank oder einer Datei gelesen werden soll
    */
    func Sichern(datei: String, alsDatenbank: Bool)
    
    /**
	Liest die Datei unter dem gegebenen Namen ein.
	- parameters:
		- datei: Dateiname
		- alsDatenbank: gibt an, ob aus einer Datenbank oder einer Datei gelesen werden soll
	- returns: wahr, wenn die Datei gelesen werden konnte
    */
    func Lesen(datei: String, alsDatenbank: Bool) -> Bool
}

/**
Kontrolliert die Anzeige des Graphen
*/
class GraphenSteuerung: EreignisEmpfänger, WerkzeugBeobachter
{
    /** Der Status für Dragging */
    private enum DraggingStatus
    {
        case keinDragging, draggingMöglich, draggingAktiv
    }

    /** Radius der Knotensymbole */
    private final let standardRadius = 20.0
    /** Farbe der Knotensymbole */
    private final let standardKnotenFarbe = "weiß"
    /** Dicke der Kantensymbole */
    private final let standardDicke = 3
    /** Farbe der Kantensymbole */
    private final let standardKantenFarbe = "blau"
    /** Die zugeordnete Oberfläche */
    private let anzeige: GraphenAnzeige
    /** Das aktive Werkzeug */
    private var aktuellesWerkzeug: Werkzeuge
    /** Der Draggingstatus */
    private var draggingStatus: DraggingStatus
    /** Die Knotensymbole */
    private var knoten: [KnotenSymbol]
    /** Die Kantensymbole */
    private var kanten: [KantenSymbol]
    /** Ein gegebenenfalls selektierter Knoten */
    private var selektierterKnoten: KnotenSymbol?
    /** Der selektierte Startknoten für eine Kante */
    private var startKnoten: KnotenSymbol?
    /** Der selektierte Zielknoten für eine Kante */
    private var zielKnoten: KnotenSymbol?
    /** Eine geklickte Kante */
    private var geklickteKante: KantenSymbol?
    /** Eine gegebenenfalls selektierte Kante */
    private var selektierteKante: KantenSymbol?
    /** Das Symbol für einen gezogenen Knoten */
    private var gezogenerKnoten: GezogenerKnotenSymbol?
    /** wenn wahr, wurde der Graph geändert. */
    private var geändert: Bool
    /** Zähler für eindeutigen Knotenbezeichner */
    private var zähler: Int
    /** x-Koordinate von Maus gedrückt */
    private var xGedrückt: CGFloat
    /** y-Koordinate von Maus gedrückt */
    private var yGedrückt: CGFloat

    /**
	Legt die Attribute an
	- parameters:
		- a: Die zugeordnete View
    */
    init(a: GraphenAnzeige)
    {
        anzeige = a
        aktuellesWerkzeug = .verschieben
        draggingStatus = .keinDragging
        knoten = [KnotenSymbol]()
        kanten = [KantenSymbol]()
        selektierterKnoten = nil
        startKnoten = nil
        zielKnoten = nil
        selektierteKante = nil
        gezogenerKnoten = nil
        geklickteKante = nil
        selektierteKante = nil
        geändert = false
        zähler = 0
        xGedrückt = CGFloat.infinity
        yGedrückt = CGFloat.infinity
        anzeige.EmpfängerAnmelden(e: self)
    }
    
    /**
    Meldet, dass sich das aktuelle Werkzeug geändert hat.
    - parameters:
        - aktWerkzeug: das neue, aktuelle Werkzeug
    */
    func AktuellesWerkzeugsetzen(aktWerkzeug: Werkzeuge)
    {
        if aktuellesWerkzeug != aktWerkzeug
        {
            if selektierterKnoten != nil
            {
                selektierterKnoten!.FarbeSetzen(f: standardKnotenFarbe)
                selektierterKnoten = nil
            }
            if selektierteKante != nil
            {
                selektierteKante!.FarbeSetzen(f: standardKantenFarbe)
                selektierteKante = nil
            }
        }
        aktuellesWerkzeug = aktWerkzeug
    }

    /**
    Die Oberfläche wurde gedrückt.
    - parameters:
        - x: x-Koordinate des Events
        - y: y-Koordinate des Events
    */
    func TouchDown(x: CGFloat, y: CGFloat)
    {
        if selektierterKnoten != nil
        {
            selektierterKnoten!.FarbeSetzen(f: standardKnotenFarbe)
            selektierterKnoten = nil
        }
        if selektierteKante != nil
        {
            selektierteKante!.FarbeSetzen(f: standardKantenFarbe)
            selektierteKante = nil
        }
        draggingStatus = .keinDragging
        let geklickterKnoten = IstKlickAufKnoten(x: x, y: y)
        switch aktuellesWerkzeug
        {
            case .knotenErstellen:
                if geklickterKnoten == nil
                {
                	xGedrückt = x
                	yGedrückt = y
                }
            case .kanteErstellen:
                if geklickterKnoten != nil
                {
                	if startKnoten == nil
                	{
                		startKnoten = geklickterKnoten
                    }
                    else
                    {
                    	zielKnoten = geklickterKnoten
                    }
                }
                else
                {
                	if startKnoten != nil
                	{
                		startKnoten!.FarbeSetzen(f: standardKnotenFarbe)
                    	startKnoten = nil
                    }
                }
            case .verschieben:
                if geklickterKnoten != nil
                {
                	selektierterKnoten = geklickterKnoten
                	draggingStatus = .draggingMöglich
                	selektierteKante = nil
                }
                else
                {
                	selektierteKante = IstKlickAufKante(x: x, y: y)
                }
        }
    }
    
    /**
    Die Oberfläche wurde losgelassen.
    - parameters:
        - x: x-Koordinate des Events
        - y: y-Koordinate des Events
    */
    func TouchUp(x: CGFloat, y: CGFloat)
    {
        switch aktuellesWerkzeug
        {
            case .knotenErstellen:
                if xGedrückt != CGFloat.infinity
                {
                	let dx = x - xGedrückt
                	let dy = y - yGedrückt
                	if dx * dx + dy * dy <= 25
                	{
                        while !BezeichnerPrüfen(bezeichner: "K\(zähler)")
                        {
                            zähler += 1
                        }
                        let k = KnotenSymbol(x: x, y: y, radius: standardRadius, farbe: standardKnotenFarbe, bezeichner: "K\(zähler)", a: anzeige)
                        knoten.append(k)
                        geändert = true
                        zähler += 1
                    }
                }
            case .kanteErstellen:
                if zielKnoten != nil
                {
                	let dx = zielKnoten!.XGeben() - x
                	let dy = zielKnoten!.YGeben() - y
                	let r = zielKnoten!.RadiusGeben()
                	if dx * dx - dy * dy <= r * r
                	{
                        let k = KantenSymbol(start: startKnoten!, ende: zielKnoten!, gerichtet: false, gewicht: "1", breite: standardDicke, farbe: standardKantenFarbe, a: anzeige)
                        kanten.append(k)
                        geändert = true
                    
                    }
                    startKnoten!.FarbeSetzen(f: standardKnotenFarbe)
                    startKnoten = nil
                    zielKnoten = nil
                }
                else if startKnoten != nil
                {
                	let dx = startKnoten!.XGeben() - x
                	let dy = startKnoten!.YGeben() - y
                	let r = startKnoten!.RadiusGeben()
                	if dx * dx - dy * dy <= r * r
                	{
                        startKnoten!.FarbeSetzen(f: "grün")
                    }
                    else
                    {
                    	startKnoten!.FarbeSetzen(f: standardKantenFarbe)
                    	startKnoten = nil
                    }
                
                }
            case .verschieben:
                if draggingStatus == .draggingAktiv
                {
                    anzeige.Entfernen(symbol: gezogenerKnoten!)
                    gezogenerKnoten = nil
                    selektierterKnoten!.FarbeSetzen(f: standardKnotenFarbe)
                    selektierterKnoten!.PositionSetzen(x: x, y: y)
                    selektierterKnoten = nil
                    geändert = true
                }
                else
                {
                    if selektierteKante != nil
                    {
                        selektierteKante!.FarbeSetzen(f: "orange")
                    }
                    else if selektierterKnoten != nil
                    {
                        selektierterKnoten!.FarbeSetzen(f: "rot")
                    }
                }
                draggingStatus = DraggingStatus.keinDragging;
        }
    }
    
    /**
    Die Oberfläche wurde gedrückt gezogen.
    - parameters:
        - x: x-Koordinate des Events
        - y: y-Koordinate des Events
    */
    func TouchGezogen(x: CGFloat, y: CGFloat)
    {
        if aktuellesWerkzeug == .verschieben
        {
            if draggingStatus == .draggingMöglich
            {
                draggingStatus = .draggingAktiv
                gezogenerKnoten = GezogenerKnotenSymbol(x: x, y: y, radius: standardRadius, a: anzeige)
                selektierterKnoten!.FarbeSetzen(f: "gelb")
            }
            else if draggingStatus == .draggingAktiv
            {
                gezogenerKnoten!.PositionSetzen(x: x, y: y)
            }
        }
    }
    
    /**
	Testet, ob der Mausklick einen existierenden Knoten trifft
	- parameters:
		- x: x-Position des Mausklicks
		- y: y-Position des Mausklicks
	- returns: getroffenen Knoten
    */
    private func IstKlickAufKnoten(x: CGFloat, y: CGFloat) -> KnotenSymbol?
    {
        for k in knoten
        {
            let dx = x - k.XGeben()
            let dy = y - k.YGeben()
            if dx * dx + dy * dy <= standardRadius*standardRadius
            {
                return k
            }
        }
        return nil
    }

    /**
	Testet, ob der Mausklick eine existierende Kante trifft
	- parameters:
		- x: x-Position des Mausklicks
		- y: y-Position des Mausklicks
	- returns: getroffene Kante
    */
    private func IstKlickAufKante(x: CGFloat, y: CGFloat) -> KantenSymbol?
    {
        for k in kanten
        {
            let x1 = k.StartsymbolGeben().XGeben()
            let y1 = k.StartsymbolGeben().YGeben()
            let x2 = k.ZielsymbolGeben().XGeben()
            let y2 = k.ZielsymbolGeben().YGeben()
            if x1 == x2
            {
                if (abs(x - x1) <= 5.0) && ((y1 < y2) && (y1 <= y) && (y <= y2) || (y1 >= y2) && (y1 >= y) && (y >= y2))
                {
                    return k
                }
            }
            else
            {
                let m = (y2 - y1) / (x2 - x1)
                let d = abs((m * (x - x1) + y1 - y) / sqrt(m*m + 1))   //HNF
                if d <= 5.0
                {
                    if abs(m) > 1
                    {
                        if (y1 < y2) && (y1 <= y) && (y <= y2) || (y1 > y2) && (y1 >= y) && (y >= y2)
                        {
                            return k
                        }
                    }
                    else
                    {
                        if (x1 < x2) && (x1 <= x) && (x <= x2) || (x1 > x2) && (x1 >= x) && (x >= x2)
                        {
                            return k
                        }
                    }
                }
            }
        }
        return nil
    }

    /**
	Sucht das zum gegebenen Namen gehörende Knotensymbol
	- parameters:
		- name: Name des Knotens
	- returns: Referenz auf das zum gegebenen Namen gehörende Knotensymbol oder nil
     */
    private func NameSuchen(name: String) -> KnotenSymbol?
    {
        for k in knoten
        {
            if k.BezeichnerGeben() == name
            {
                return k
            }
        }
        return nil
    }

    /**
	Speichert die Datei am angegebenen Ort.

	Eine exisiterende Datei wird gegebenenfalls überschrieben, eine Datenbank zuerst gelöscht.
	- parameters:
		- datei: Dateiname
		- alsDatenbank: gibt an, ob aus einer Datenbank oder einer Datei gelesen werden soll
    */
    func Sichern(datei: String, alsDatenbank: Bool)
    {
        if alsDatenbank
        {
            SichernDatenbank(name: datei)
        }
        else
        {
            SichernDatei(datei: datei)
        }
    }
    
    /**
	Speichert die Datei am angegebenen Ort.

	Eine exisiterende Datei wird gegebenenfalls überschrieben.
	- parameters:
		- datei: Dateiname
    */
    private func SichernDatei(datei: String)
    {
    	var ausgabe = "Knoten: \(knoten.count)\n"
    	for k in knoten
    	{
    		ausgabe = ausgabe + "\(k.XGeben())\t\(k.YGeben())\t\(k.BezeichnerGeben())\n"
        }
        ausgabe = ausgabe + "Kanten: \(kanten.count)\n"
        for k in kanten
        {
        	ausgabe = ausgabe + "\(k.StartsymbolGeben().BezeichnerGeben())\t\(k.ZielsymbolGeben().BezeichnerGeben())\t\(k.GewichtGeben())\t\(k.IstGerichtet())\n"
        }
        do
        {
            let ziel = try? FileManager.default.url(for: .documentDirectory, in: .userDomainMask, appropriateFor: nil, create: true).appendingPathComponent(datei)
            if FileManager.default.fileExists(atPath: ziel!.relativePath)
            {
                try? FileManager.default.removeItem(at: ziel!)
            }
            FileManager.default.createFile(atPath: ziel!.relativePath, contents: nil)
            try ausgabe.write(to: ziel!, atomically: true, encoding: .utf8)
            geändert = false
        }
        catch
        {
        	print("Fehler beim Schreiben der Datei '\(datei)': \(error)")
        }
    }

    /**
	Speichert den Graphen als Datenbank in der angegebenen Datei.
	
	Eine exisiterende Datenbank wird gegebenenfalls überschrieben.
	- parameters:
		- name: Name der Datenbankcontainerdatei
    */
    private func SichernDatenbank(name: String)
    {
        do
        {
            let verbindung = try DatenbankVerbindung(name: name)
            var anweisung = try verbindung.AnweisungErzeugen(sql: "DROP TABLE IF EXISTS knoten")
            try anweisung.AnweisungAusführen()
            anweisung.AnweisungSchliessen()
            anweisung = try verbindung.AnweisungErzeugen(sql: "DROP TABLE IF EXISTS kanten")
            try anweisung.AnweisungAusführen()
            anweisung.AnweisungSchliessen()
            anweisung = try verbindung.AnweisungErzeugen(sql: "CREATE TABLE \"knoten\" (\"bezeichner\" TEXT,\"x\" INTEGER NOT NULL, \"y\" INTEGER NOT NULL, PRIMARY KEY(\"bezeichner\"))")
            try anweisung.AnweisungAusführen()
            anweisung.AnweisungSchliessen()
            anweisung = try verbindung.AnweisungErzeugen(sql: "CREATE TABLE \"kanten\" (\"bezeichnerStart\" TEXT, \"bezeichnerZiel\" TEXT, \"gewicht\" TEXT, \"gerichtet\" TEXT)")
            try anweisung.AnweisungAusführen()
            anweisung.AnweisungSchliessen()

            for k in  knoten
            {
                anweisung = try verbindung.AnweisungErzeugen(sql: "INSERT INTO knoten (bezeichner, x, y) VALUES ('\(k.BezeichnerGeben())', \(k.XGeben()), \(k.YGeben()))")
                try anweisung.AnweisungAusführen()
                anweisung.AnweisungSchliessen()
            }
            for k in kanten
            {
                anweisung = try verbindung.AnweisungErzeugen(sql: "INSERT INTO kanten (bezeichnerStart, bezeichnerZiel, gewicht, gerichtet) VALUES ('\(k.StartsymbolGeben().BezeichnerGeben())', '\(k.ZielsymbolGeben().BezeichnerGeben())', \(k.GewichtGeben()), \(k.IstGerichtet()))")
                try anweisung.AnweisungAusführen()
                anweisung.AnweisungSchliessen()
            }
            verbindung.VerbindungSchliessen()
            geändert = false
        }
        catch
        {
            if let dberror = error as? SQLiteFehler
            {
            	print ("SQLite-Fehler beim Bearbeiten der Datenbank: \(dberror.localizedDescription)")
            }
            else
            {
                print("Unbekannter Fehler beim Lesen der Datenbank: \(error)")
            }
        }
    }

    /**
	Liest die Datei unter dem gegebenen Namen ein.
	- parameters:
		- datei: Dateiname
		- alsDatenbank: gibt an, ob aus einer Datenbank oder einer Datei gelesen werden soll
	- returns: wahr, wenn die Datei gelesen werden konnte
    */
    func Lesen(datei: String, alsDatenbank: Bool) -> Bool
    {
        ZurückSetzen()
        if alsDatenbank
        {
            return LesenDatenbank(name: datei)
        }
        else
        {
            return LesenDatei(datei: datei)
        }
    }

    /**
	Liest die Datei unter dem gegebenen Namen ein.
	- parameters:
		- datei: Dateiname
	- returns: wahr, wenn die Datei gelesen werden konnte
    */
    private func LesenDatei(datei: String) -> Bool
    {
        do
        {
            let ziel = try? FileManager.default.url(for: .documentDirectory, in: .userDomainMask, appropriateFor: nil, create: true).appendingPathComponent(datei)
            let inhalt = try String(contentsOf: ziel!)
            let zeilen = inhalt.split(separator: "\n")
            var teile = zeilen[0].split(separator: ": ")
            let anzahl = Int(teile[1])!
            for i in 1 ... anzahl
            {
            	teile = zeilen[i].split(separator: "\t")
            	let x = Double(teile[0])
            	let y = Double(teile[1])
                knoten.append(KnotenSymbol(x: x!, y: y!, radius: standardRadius, farbe: standardKnotenFarbe, bezeichner: String(teile[2]), a: anzeige))
            }
            teile = zeilen[anzahl + 1].split(separator: ": ")
            let anzahl2 = Int(teile[1])!
            for i in anzahl + 2 ... anzahl + anzahl2 + 1
            {
            	teile = zeilen[i].split(separator: "\t")
                kanten.append(KantenSymbol(start: NameSuchen(name: String(teile[0]))!, ende: NameSuchen(name: String(teile[1]))!, gerichtet: teile[3] == "true", gewicht: String(teile[2]), breite: standardDicke, farbe: standardKantenFarbe, a: anzeige))
            }
            return true
        }
        catch
        {
			print("Fehler beim Lesen der Datei '" + datei + "': " + error.localizedDescription)
        	return false
        }
    }

    /**
	Liest die Datenbank unter dem gegebenen Namen ein.
	- parameters:
		- name: Name der Datenbankcontainerdatei
        - returns: wahr, wenn die Datenbank erfolgreich gelesen werden konnte
     */
    func LesenDatenbank(name: String) -> Bool
    {
        ZurückSetzen()
        do
        {
            let verbindung = try DatenbankVerbindung(name: name)
            var anweisung = try verbindung.AnweisungErzeugen(sql: "SELECT bezeichner, x, y FROM knoten")
            var daten = try anweisung.AbfrageAusführen()
            while daten.AufNächstenDatensatzPositionieren()
            {

                try knoten.append(KnotenSymbol(x: CGFloat(daten.SpaltenwertGebenInt(spaltenname: "x")), y: CGFloat(daten.SpaltenwertGebenInt(spaltenname: "y")), radius: standardRadius, farbe: standardKnotenFarbe, bezeichner: daten.SpaltenwertGebenString(spaltenname: "bezeichner"), a: anzeige))

            }
            anweisung.AnweisungSchliessen()
            anweisung = try verbindung.AnweisungErzeugen(sql: "SELECT bezeichnerStart, bezeichnerZiel, gewicht, gerichtet FROM kanten")
            daten = try anweisung.AbfrageAusführen()
            while daten.AufNächstenDatensatzPositionieren()
            {
                try kanten.append(KantenSymbol(start: NameSuchen(name: daten.SpaltenwertGebenString(spaltenname: "bezeichnerStart"))!, ende: NameSuchen(name: daten.SpaltenwertGebenString(spaltenname: "bezeichnerZiel"))!, gerichtet: daten.SpaltenwertGebenString(spaltenname: "gerichtet") != "0", gewicht: daten.SpaltenwertGebenString(spaltenname: "gewicht"), breite: standardDicke, farbe: standardKantenFarbe, a: anzeige))
            }
            anweisung.AnweisungSchliessen()
            verbindung.VerbindungSchliessen()
            return true
        }
        catch
        {
            if let dberror = error as? SQLiteFehler
            {
            	print ("SQLite-Fehler beim Bearbeiten der Datenbank: \(dberror.localizedDescription)")
            }
            else
            {
                print("Unbekannter Fehler beim Lesen der Datenbank: \(error)")
            }
            return false
        }
    }

    /**
	Gibt die SQL-Statements für das Speichern des Graphen auf Konsole aus
    */
    func SichernKonsole()
    {
        print("DROP TABLE IF EXISTS knoten")
        print("DROP TABLE IF EXISTS kanten")
        print("CREATE TABLE \"knoten\" (\"bezeichner\" TEXT,\"x\" INTEGER NOT NULL, \"y\" INTEGER NOT NULL, PRIMARY KEY(\"bezeichner\"))")
        print("CREATE TABLE \"kanten\" (\"bezeichnerStart\" TEXT, \"bezeichnerZiel\" TEXT, \"gewicht\" TEXT, \"gerichtet\" TEXT)")
        for k in  knoten
        {
            print("INSERT INTO knoten (bezeichner, x, y) VALUES ('\(k.BezeichnerGeben())', \(k.XGeben()), \(k.YGeben()))")
        }
        for k in kanten
        {
            print("INSERT INTO kanten (bezeichnerStart, bezeichnerZiel, gewicht, gerichtet) VALUES ('\(k.StartsymbolGeben().BezeichnerGeben())', '\(k.ZielsymbolGeben().BezeichnerGeben())', \(k.GewichtGeben()), \(k.IstGerichtet()))")
        }
    }

    /**
	Prüft, ob ein Knotenbezeichner eindeutig ist.
	
	Der selektierte Knoten wird nicht betrachtet, da sein Name bleiben kann.
	- parameters:
		- bezeichner: der zu prüfende Bezeichner
	- returns: wahr, wenn der Bezeichner möglich ist
     */
    func BezeichnerPrüfen(bezeichner: String) -> Bool
    {
        for k in knoten
        {
            if (k !== selektierterKnoten) && (k.BezeichnerGeben() == bezeichner)
            {
                return false
            }
        }
        return true
    }
    
    /**
    Löscht das aktuell ausgewählte Objekt
     */
    func Löschen()
    {
        if selektierterKnoten != nil
        {
            knoten.removeAll(where: {$0 === selektierterKnoten})
            anzeige.Entfernen(symbol: selektierterKnoten!)
            let ka = kanten
            for k in ka
            {
                if (k.StartsymbolGeben() === selektierterKnoten) || (k.ZielsymbolGeben() === selektierterKnoten)
                {
                    kanten.removeAll(where: {$0 === k})
                    anzeige.Entfernen(symbol: k)
                }
            }
            selektierterKnoten = nil
            geändert = true
        }
        else if selektierteKante != nil
        {
            kanten.removeAll(where: {$0 === selektierteKante})
            anzeige.Entfernen(symbol: selektierteKante!)
            selektierteKante = nil
            geändert = true
        }
    }
    
    /**
    Meldet zurück, ob das Szenario verändert wurde.
	- returns: wahr, wenn das Szenario verändert wurde
     */
    func IstGeändert() -> Bool
    {
        return geändert
    }
    
    /**
    Meldet den selektierten Knoten zurück
	- returns: selektierter Knoten oder nil
    */
    func SelektiertenKnotenGeben() -> KnotenSymbol?
    {
    	return selektierterKnoten
    }
    
    /**
    Meldet die selektierte Kante zurück
	- returns: selektierte Kante oder nil
    */
    func SelektierteKanteGeben() -> KantenSymbol?
    {
    	return selektierteKante
    }
    
    /**
    Löscht das Szenarium
    */
    func ZurückSetzen()
    {
        for k in knoten
        {
            anzeige.Entfernen(symbol: k)
        }
        knoten.removeAll()
        for k in kanten
        {
            anzeige.Entfernen(symbol: k)
        }
        kanten.removeAll()
        selektierterKnoten = nil
        startKnoten = nil
        zielKnoten = nil
        selektierteKante = nil
        geändert = false
    }
}

/**
Zeigt den Hilfetext an
*/
class HilfeAnzeige: UITextView
{
    /** Kontrolleur für die Gesamtoberfläche */
    private var oberfläche: OberflächenController! = nil
    /** Startzeit für den Touch um  Klichverhalten zu ermitteln */
    private var klickZeit: TimeInterval?
    /** Der Hilfetext */
    private let hilfetext =
"""
Grundlegende Funktionen
- Neuen Graph anlegen
- Gespeicherten Graph öffnen
- Aktuellen Graph sichern
- Aktuellen Graph exportieren
- Diese Hilfe anzeigen
Alle sechs Grundfunktionen können über die entsprechenden Bedienknöpfe auf der rechten Seite der Werkzeugleiste ausgeführt werden.

Auswahlmodus einschalten
Auf das Pfeilsymbol am linken Rand der Werkzeugleiste klicken

Neuen Knoten einfügen
Knotensymbol im linken Teil der Werkzeugleiste auswählen und mit der Maus auf die Position für den neuen KLnoten klicken.

Knoten verschieben
Auswahlmodus in der Werkzeugleiste auswählen und Knoten mit der Maus an den gewünschten Ort ziehen;
alle Kanten werden mit gezogen

Knoten editieren
Auswahlmodus in der Werkzeugleiste auswählen, Knoten anklicken und im rechten Teil der Werkzeugleiste auf das Editieren-Symbol (angedeuteter Dialog) klicken

Knoten löschen
Auswahlmodus in der Werkzeugleiste auswählen, Knoten anklicken und im rechten Teil der Werkzeugleiste auf das Löschen-Symbol (Mülleimer) klicken

Neue Kante einfügen
Kantensymbol in der Werkzeugleiste auswählen und mit der Maus vom Sartknoten zum gewünschten Zielknoten ziehen

Kante editieren
Auswahlmodus in der Werkzeugleiste auswählen, Kante anklicken und im rechten Teil der Werkzeugleiste auf das Editieren-Symbol (angedeuteter Dialog) klicken

Kante löschen
Auswahlmodus in der Werkzeugleiste auswählen, Kante anklicken und im rechten Teil der Werkzeugleiste auf das Löschen-Symbol (Mülleimer) klicken

                                   Zum Beenden der Hilfeanzeige kurz klicken
"""

    /**
    Standardkonstruktor, legt die Größe und Position fest
    - parameters:
        - frame: Framerechteck
        - oberfläche: Oberflächenkontroller zum Zurücksdchslten auf Normalansicht
    */
    init(frame: CGRect, oberfläche: OberflächenController)
    {
        klickZeit = 0
        self.oberfläche = oberfläche
        super.init(frame: frame, textContainer: nil)
        text = hilfetext
        isEditable = false
        isSelectable = false
    }

    /**
    Standardkonstruktor, liest daten über den Decoder
    - parameters:
        - coder: Dekodierobjekt
    */
    required init?(coder: NSCoder)
    {
        klickZeit = 0
        super.init(coder: coder)
        text = hilfetext
        isEditable = false
        isSelectable = false
    }
    
    /**
    Aufgerufen, wenn ein oder mehr Finger die Tastatur berühren.
    - parameters:
        - touches: Menge an Berüchrungsbeschreibungen für den Beginn des Ereignisses
        - event: das Gesamtereignis, zu dem die Berührungen gehören
    */
    override func touchesBegan(_ touches: Set<UITouch>, with event: UIEvent?) {
        klickZeit = event?.timestamp
        super.touchesBegan(touches, with: event)
    }

    /**
    Aufgerufen, wenn ein oder mehr Finger die Tastatur verlassen.
    - parameters:
        - touches: Menge an Berührungsbeschreibungen für den Beginn des Ereignisses
        - event: das Gesamtereignis, zu dem die Berührungen gehören
    */
    override func touchesEnded(_ touches: Set<UITouch>, with event: UIEvent?)
    {
        var ok = false
        let zeit = event?.timestamp
        if (zeit == nil) || (klickZeit == nil) || (zeit! - klickZeit! < 0.2)
        {
            ok = true
        }
        super.touchesEnded(touches, with: event)
        if ok
        {
            oberfläche.HilfeRücksetzen()
        }
    }
}

/**
Dialog für das Öffnen und Speichern von Dateien
*/
class DateiDialog: UIView
{
    /** Kontrolleur für die Gesamtoberfläche */
    private var oberfläche: OberflächenController! = nil
    /** Überschrift */
    private var überschrift: UILabel
    /** Dateiname */
    private var dateiname: UITextField
    /** Als Datenbank */
    private let alsDatenbank: UISwitch
    /** OK-Knopf */
    private var ok: UIButton
    /** Abbrechen-Knopf */
    private var abbrechen: UIButton

    /**
    Standardkonstruktor, legt die Größe und Position fest
    - parameters:
        - frame: Framerechteck
        - oberfläche: Oberflächenkontroller zum Zurücksdchslten auf Normalansicht
    */
    init(frame: CGRect, oberfläche: OberflächenController)
    {
        self.oberfläche = oberfläche
        überschrift = UILabel(frame: CGRect(x: 0.0, y: 0.0, width: frame.width, height: 40.0))
        überschrift.text = "Datei öffnen"
        überschrift.textAlignment = .center
        dateiname = UITextField(frame: CGRect(x: frame.width / 2 - 100, y: 50.0, width: 200, height: 60.0))
        dateiname.text = ""
        dateiname.placeholder = "Dateinamen eingeben"
        dateiname.textAlignment = .center
        alsDatenbank = UISwitch(frame: CGRect(x: frame.width / 2 + 60, y: 120, width: 20, height: 40))
        alsDatenbank.setOn(true, animated: false)
        alsDatenbank.preferredStyle = .checkbox
        let alsDatenbankLabel = UILabel(frame: CGRect(x: frame.width / 2 - 80, y: 115, width: 140, height: 40))
        alsDatenbankLabel.text = "Als Datenbank: "
        alsDatenbankLabel.textAlignment = .right
        ok = UIButton(type: .system)
        abbrechen = UIButton(type: .system)
        super.init(frame: frame)
        ok.setTitle("OK", for: .normal)
        ok.isEnabled = true
        ok.frame = CGRect(x: frame.width / 2 + 20, y: 170, width: 80, height: 40)
        ok.role = .primary
        ok.addTarget(self, action: #selector(OkGedrückt), for: .touchUpInside)
        abbrechen.setTitle("Abbrechen", for: .normal)
        abbrechen.isEnabled = true
        abbrechen.frame = CGRect(x: frame.width / 2 - 100, y: 170, width: 80, height: 40)
        abbrechen.role = .cancel
        abbrechen.addTarget(self, action: #selector(AbbrechenGedrückt), for: .touchUpInside)
        self.addSubview(überschrift)
        self.addSubview(dateiname)
        self.addSubview(alsDatenbank)
        self.addSubview(alsDatenbankLabel)
        self.addSubview(ok)
        self.addSubview(abbrechen)
    }

    /**
    Standardkonstruktor, liest daten über den Decoder
    - parameters:
        - coder: Dekodierobjekt
    */
    required init?(coder: NSCoder)
    {
        überschrift = UILabel(coder: coder)!
        überschrift.text = "Datei öffnen"
        überschrift.textAlignment = .center
        dateiname = UITextField(coder: coder)!
        dateiname.text = ""
        dateiname.placeholder = "Dateinamen eingeben"
        alsDatenbank = UISwitch(coder: coder)!
        ok = UIButton(frame: CGRect(x: 100, y: 100, width: 80, height: 40))
        abbrechen = UIButton(frame: CGRect(x: 0, y: 100, width: 80, height: 40))
        super.init(coder: coder)
    }
    
    /**
    	Überschrift anpassen
    */
    func ÜberschriftSetzen (sichern: Bool)
    {
    	überschrift.text = sichern ? "Speichern auf Datei" : "Datei öffnen"
    }
    
    /**
	Ok gedrückt
    */
	@objc func OkGedrückt()
	{
		if var text = dateiname.text
  		{
        	if !text.isEmpty
        	{
        		if alsDatenbank.isOn
            	{
                    if !text.description.hasSuffix(".grdb")
                    {
                        text += ".grdb"
                    }
                }
                else
                {
                    if !text.description.hasSuffix(".grph")
                    {
                        text += ".grph"
                    }
                }
                oberfläche.DateidialogOK(name: text, alsDatenbank: alsDatenbank.isOn)
            }
        }
    }
    
    /**
	Cancel gedrückt
    */
	@objc func AbbrechenGedrückt()
	{
        oberfläche.DateidialogAbbrechen()
    }
}

/**
Dialog für das Editieren eines Knotens
*/
class KnotenDialog: UIView
{
    /** Kontrolleur für die Gesamtoberfläche */
    private var oberfläche: OberflächenController! = nil
    /** Überschrift */
    private var überschrift: UILabel
    /** Überschrift */
    private var bezeichnerLab: UILabel
    /** Überschrift */
    private var xLab: UILabel
    /** Überschrift */
    private var yLab: UILabel
    /** Knotenbezeichner */
    private var bezeichner: UITextField
    /** X-Koordinate */
    private var x: UITextField
    /** Y-Koordinate */
    private var y: UITextField
    /** OK-Knopf */
    private var ok: UIButton
    /** Abbrechen-Knopf */
    private var abbrechen: UIButton

    /**
    Standardkonstruktor, legt die Größe und Position fest
    - parameters:
        - frame: Framerechteck
        - oberfläche: Oberflächenkontroller zum Zurücksdchslten auf Normalansicht
    */
    init(frame: CGRect, oberfläche: OberflächenController)
    {
        self.oberfläche = oberfläche
        überschrift = UILabel(frame: CGRect(x: 0.0, y: 0.0, width: frame.width, height: 40.0))
        überschrift.text = "Knoten editieren"
        überschrift.textAlignment = .center
        bezeichnerLab = UILabel(frame: CGRect(x: 0.0, y: 55.0, width: 100.0, height: 40.0))
        bezeichnerLab.text = "Bezeichner:"
        bezeichnerLab.textAlignment = .natural
        xLab = UILabel(frame: CGRect(x: 0.0, y: 105.0, width: 100.0, height: 40.0))
        xLab.text = "X-Koordinate:"
        xLab.textAlignment = .natural
        yLab = UILabel(frame: CGRect(x: 0.0, y: 155.0, width: 100.0, height: 40.0))
        yLab.text = "Y-Koordinate:"
        yLab.textAlignment = .natural
        bezeichner = UITextField(frame: CGRect(x: 110.0, y: 50.0, width: 200.0, height: 50.0))
        bezeichner.text = ""
        bezeichner.placeholder = "Bezeichner"
        bezeichner.textAlignment = .natural
        x = UITextField(frame: CGRect(x: 110.0, y: 100.0, width: 200.0, height: 50.0))
        x.text = ""
        x.placeholder = "X-Koordinate"
        x.textAlignment = .natural
        y = UITextField(frame: CGRect(x: 110.0, y: 150.0, width: 200.0, height: 50.0))
        y.text = ""
        y.placeholder = "Y-Koordinate"
        y.textAlignment = .natural
        ok = UIButton(type: .system)
        abbrechen = UIButton(type: .system)
        super.init(frame: frame)
        ok.setTitle("OK", for: .normal)
        ok.isEnabled = true
        ok.frame = CGRect(x: frame.width / 2 + 20, y: 170, width: 80, height: 40)
        ok.role = .primary
        ok.addTarget(self, action: #selector(OkGedrückt), for: .touchUpInside)
        abbrechen.setTitle("Abbrechen", for: .normal)
        abbrechen.isEnabled = true
        abbrechen.frame = CGRect(x: frame.width / 2 - 100, y: 170, width: 80, height: 40)
        abbrechen.role = .cancel
        abbrechen.addTarget(self, action: #selector(AbbrechenGedrückt), for: .touchUpInside)
        self.addSubview(überschrift)
        self.addSubview(bezeichnerLab)
        self.addSubview(xLab)
        self.addSubview(yLab)
        self.addSubview(bezeichner)
        self.addSubview(x)
        self.addSubview(y)
        self.addSubview(ok)
        self.addSubview(abbrechen)
    }

    /**
    Standardkonstruktor, liest daten über den Decoder
    - parameters:
        - coder: Dekodierobjekt
    */
    required init?(coder: NSCoder)
    {
        überschrift = UILabel(coder: coder)!
        bezeichnerLab = UILabel(coder: coder)!
        xLab = UILabel(coder: coder)!
        yLab = UILabel(coder: coder)!
        bezeichner = UITextField(coder: coder)!
        x = UITextField(coder: coder)!
        y = UITextField(coder: coder)!
        ok = UIButton(frame: CGRect(x: 100, y: 100, width: 80, height: 40))
        abbrechen = UIButton(frame: CGRect(x: 0, y: 100, width: 80, height: 40))
        super.init(coder: coder)
    }
    
    /**
    Daten setzen
	- parameters:
		- bezeichner: der bisherige Knotenbezeichner
    	- x: die bisherige x-Koordinate
    	- y: die bisherige y-Koordinate
    */
    func DatenSetzen (bezeichner: String, x: CGFloat, y: CGFloat)
    {
    	self.bezeichner.text = bezeichner
    	self.x.text = "\(x)"
    	self.y.text = "\(y)"
    }
    
    /**
	Ok gedrückt
    */
	@objc func OkGedrückt()
	{
        if let text = bezeichner.text, let x = Double(self.x.text ?? ""), let y = Double(self.y.text ?? "")
  		{
            oberfläche.KnotendialogOK(bezeichner: text, x: x, y: y)
        }
    }
    
    /**
	Cancel gedrückt
    */
	@objc func AbbrechenGedrückt()
	{
        oberfläche.KnotendialogAbbrechen()
    }
}

/**
Dialog für das Editieren einer Kante
*/
class KantenDialog: UIView
{
    /** Kontrolleur für die Gesamtoberfläche */
    private var oberfläche: OberflächenController! = nil
    /** Überschrift */
    private var überschrift: UILabel
    /** Markierung Gewicht */
    private var gewichtLab: UILabel
    /** Markierung gerichtet */
    private var gerichtetLab: UILabel
    /** Gewicht */
    private var gewicht: UITextField
    /** gerichtet */
    private var gerichtet: UISwitch
    /** OK-Knopf */
    private var ok: UIButton
    /** Abbrechen-Knopf */
    private var abbrechen: UIButton

    /**
    Standardkonstruktor, legt die Größe und Position fest
    - parameters:
        - frame: Framerechteck
        - oberfläche: Oberflächenkontroller zum Zurücksdchslten auf Normalansicht
    */
    init(frame: CGRect, oberfläche: OberflächenController)
    {
        self.oberfläche = oberfläche
        überschrift = UILabel(frame: CGRect(x: 0.0, y: 0.0, width: frame.width, height: 40.0))
        überschrift.text = "Kante editieren"
        überschrift.textAlignment = .center
        gewichtLab = UILabel(frame: CGRect(x: 0.0, y: 55.0, width: 100.0, height: 40.0))
        gewichtLab.text = "Gewicht:"
        gewichtLab.textAlignment = .natural
        gerichtetLab = UILabel(frame: CGRect(x: 0.0, y: 105.0, width: 100.0, height: 40.0))
        gerichtetLab.text = "Gerichtet:"
        gerichtetLab.textAlignment = .natural
        gewicht = UITextField(frame: CGRect(x: 110.0, y: 50.0, width: 200.0, height: 50.0))
        gewicht.text = ""
        gewicht.placeholder = "Kantengewicht"
        gewicht.textAlignment = .natural
        gerichtet = UISwitch(frame: CGRect(x: 110.0, y: 100.0, width: 200.0, height: 50.0))
        gerichtet.setOn(true, animated: false)
        gerichtet.preferredStyle = .checkbox
        gerichtet.title = "Kantengewicht"
        ok = UIButton(type: .system)
        abbrechen = UIButton(type: .system)
        super.init(frame: frame)
        ok.setTitle("OK", for: .normal)
        ok.isEnabled = true
        ok.frame = CGRect(x: frame.width / 2 + 20, y: 120, width: 80, height: 40)
        ok.role = .primary
        ok.addTarget(self, action: #selector(OkGedrückt), for: .touchUpInside)
        abbrechen.setTitle("Abbrechen", for: .normal)
        abbrechen.isEnabled = true
        abbrechen.frame = CGRect(x: frame.width / 2 - 100, y: 120, width: 80, height: 40)
        abbrechen.role = .cancel
        abbrechen.addTarget(self, action: #selector(AbbrechenGedrückt), for: .touchUpInside)
        self.addSubview(überschrift)
        self.addSubview(gewichtLab)
        self.addSubview(gerichtetLab)
        self.addSubview(gewicht)
        self.addSubview(gerichtet)
        self.addSubview(ok)
        self.addSubview(abbrechen)
    }

    /**
    Standardkonstruktor, liest daten über den Decoder
    - parameters:
        - coder: Dekodierobjekt
    */
    required init?(coder: NSCoder)
    {
        überschrift = UILabel(coder: coder)!
        gewichtLab = UILabel(coder: coder)!
        gerichtetLab = UILabel(coder: coder)!
        gewicht = UITextField(coder: coder)!
        gerichtet = UISwitch(coder: coder)!
        ok = UIButton(frame: CGRect(x: 100, y: 100, width: 80, height: 40))
        abbrechen = UIButton(frame: CGRect(x: 0, y: 100, width: 80, height: 40))
        super.init(coder: coder)
    }
    
    /**
    Daten setzen
	- parameters:
		- gewicht: das bisherige Kantengewicht
    	- gerichtet: der bisherige Wert für gerichtet
    */
    func DatenSetzen (gewicht: String, gerichtet: Bool)
    {
    	self.gewicht.text = gewicht
    	self.gerichtet.setOn(gerichtet, animated: false)
    }
    
    /**
	Ok gedrückt
    */
	@objc func OkGedrückt()
	{
        if let text = gewicht.text
  		{
            oberfläche.KantendialogOK(gewicht: text, gerichtet: gerichtet.isOn)
        }
    }
    
    /**
	Cancel gedrückt
    */
	@objc func AbbrechenGedrückt()
	{
        oberfläche.KantendialogAbbrechen()
    }
}

/**
Dialog für das Abfragen bei ungesicherten Änderungen
*/
class AbfrageDialog: UIView
{
    /** Kontrolleur für die Gesamtoberfläche */
    private var oberfläche: OberflächenController! = nil
    /** Überschrift */
    private var überschrift: UILabel
    /** Ja-Knopf */
    private var ja: UIButton
    /** Nein-Knopf */
    private var nein: UIButton

    /**
    Standardkonstruktor, legt die Größe und Position fest
    - parameters:
        - frame: Framerechteck
        - oberfläche: Oberflächenkontroller zum Zurücksdchslten auf Normalansicht
    */
    init(frame: CGRect, oberfläche: OberflächenController)
    {
        self.oberfläche = oberfläche
        überschrift = UILabel(frame: CGRect(x: 0.0, y: 0.0, width: frame.width, height: 40.0))
        überschrift.text = "Ungesicherte Änderungen verwerfen?"
        überschrift.textAlignment = .center
        ja = UIButton(type: .system)
        nein = UIButton(type: .system)
        super.init(frame: frame)
        ja.setTitle("Ja", for: .normal)
        ja.isEnabled = true
        ja.frame = CGRect(x: frame.width / 2 + 20, y: 50, width: 80, height: 40)
        ja.role = .primary
        ja.addTarget(self, action: #selector(JaGedrückt), for: .touchUpInside)
        nein.setTitle("Nein", for: .normal)
        nein.isEnabled = true
        nein.frame = CGRect(x: frame.width / 2 - 100, y: 50, width: 80, height: 40)
        nein.role = .cancel
        nein.addTarget(self, action: #selector(NeinGedrückt), for: .touchUpInside)
        self.addSubview(überschrift)
        self.addSubview(ja)
        self.addSubview(nein)
    }

    /**
    Standardkonstruktor, liest daten über den Decoder
    - parameters:
        - coder: Dekodierobjekt
    */
    required init?(coder: NSCoder)
    {
        überschrift = UILabel(coder: coder)!
        überschrift.text = "Ungesicherte Änderungen verwerfen?"
        überschrift.textAlignment = .center
        ja = UIButton(frame: CGRect(x: 100, y: 100, width: 80, height: 40))
        nein = UIButton(frame: CGRect(x: 0, y: 100, width: 80, height: 40))
        super.init(coder: coder)
    }
    
    /**
	Ja gedrückt
    */
	@objc func JaGedrückt()
	{
        oberfläche.AbfragedialogJa()
    }
    
    /**
	Nein gedrückt
    */
	@objc func NeinGedrückt()
	{
        oberfläche.AbfragedialogNein()
    }
}


/**
Kontrolliert die Hauptview
*/
class OberflächenController : UIViewController
{
    /** Der Empfänger für die Oberflächenereignisse */
    private var empfänger: EreignisEmpfänger! = nil
    /** Anzeigeview für die Schaltung */
    private var graph: GraphenAnzeige! = nil
    /** Anzeigeview für die Werkzeugleiste */
    private var werkzeugleiste: Werkzeugleiste! = nil
    /** Anzeigeview für die Werkzeugleiste */
    private var steuerung: GraphenSteuerung! = nil
    /** Anzeigeview für den Hilfetext */
    private var hilfe: HilfeAnzeige! = nil
    /** Anzeigeview für den Abfragedialog */
    private var abfrage: AbfrageDialog! = nil
    /** Anzeigeview für den Dateinamensdialog */
    private var dateidialog: DateiDialog! = nil
    /** Name der aktuellen Datei */
    private var aktuelleDatei: String = ""
    /** Aktuelle Datei ist Datenbank*/
    private var aktuellIstDatenbank: Bool = true
    /** Dateidialog zum Sichern oder Lesen */
    private var willSichern: Bool = false
    /** Aktion bei positiver Antwort des Abfragedialogs */
    private var abfrageAktion: () -> Void = {}
    /** Anzeigeview für den Knoteneditierdialog */
    private var knotenEditieren: KnotenDialog! = nil
    /** Anzeigeview für den Kanteneditierdialog */
    private var kantenEditieren: KantenDialog! = nil
    

    override func loadView() {
        let view = UIView(frame: CGRect(x: 0, y: 0, width: 700, height: 450))
        view.backgroundColor = .white

        graph = GraphenAnzeige(frame: CGRect(x: 0, y: 0, width: 700, height: 400))
        view.addSubview(graph)
        
        werkzeugleiste = Werkzeugleiste(frame: CGRect(x: 0, y: 400, width: 700, height: 50))
        view.addSubview(werkzeugleiste)
        
        hilfe = HilfeAnzeige(frame: CGRect(x: 0, y: 0, width: 700, height: 450),oberfläche: self)
        hilfe.isHidden = true
        view.addSubview(hilfe)
        
        dateidialog = DateiDialog(frame: CGRect(x: 0, y: 0, width: 700, height: 450), oberfläche: self)
        dateidialog.isHidden = true
        view.addSubview(dateidialog)
        
        abfrage = AbfrageDialog(frame: CGRect(x: 0, y: 0, width: 700, height: 450), oberfläche: self)
        abfrage.isHidden = true
        view.addSubview(abfrage)
        
        knotenEditieren = KnotenDialog(frame: CGRect(x: 0, y: 0, width: 700, height: 450), oberfläche: self)
        knotenEditieren.isHidden = true
        view.addSubview(knotenEditieren)
        
        kantenEditieren = KantenDialog(frame: CGRect(x: 0, y: 0, width: 700, height: 450), oberfläche: self)
        kantenEditieren.isHidden = true
        view.addSubview(kantenEditieren)
        
        
        self.view = view
        
        steuerung = GraphenSteuerung(a: graph)
        werkzeugleiste.OberflächeSetzen(o: self)
        werkzeugleiste.BeobachterAnmelden(b: steuerung)
        empfänger = steuerung
    }

	/**
	Zeigt die Normalansicht mit Graph und Werkzeugleiste
    */
    private func NormalansichtZeigen()
    {
        hilfe.isHidden = true
        graph.isHidden = false
        werkzeugleiste.isHidden = false
        dateidialog.isHidden = true
        abfrage.isHidden = true
        knotenEditieren.isHidden = true
        kantenEditieren.isHidden = true
    }

    /**
    Ausgeben des Graphen auf Konsole
    */
    func AusgebenAusführen()
    {
    	//steuerung.SichernKonsole()
    	SichernAusführen(immer: false)
    	if !aktuelleDatei.isEmpty
    	{
    		let ziel = try? FileManager.default.url(for: .documentDirectory, in: .userDomainMask, appropriateFor: nil, create: true).appendingPathComponent(aktuelleDatei)
    		let picker = UIDocumentPickerViewController(forExporting: [ziel!], asCopy: true)
    		present(picker, animated: true)
        }
    }

    /**
    Zeigt die Hilfe-Ansicht
    */
    func HilfeAusführen()
    {
        graph.isHidden = true
        werkzeugleiste.isHidden = true
        hilfe.isHidden = false
        dateidialog.isHidden = true
        abfrage.isHidden = true
        knotenEditieren.isHidden = true
        kantenEditieren.isHidden = true
    }

    /**
    Verbirgt die Hilfe-Ansicht
    */
    func HilfeRücksetzen()
    {
    	NormalansichtZeigen()
    }
    
    /**
	Entfernt den Dateilesen-Dialog
    */
    func DateidialogOK(name: String, alsDatenbank: Bool)
    {
        hilfe.isHidden = true
        graph.isHidden = false
        werkzeugleiste.isHidden = false
        dateidialog.isHidden = true
        abfrage.isHidden = true
        knotenEditieren.isHidden = true
        kantenEditieren.isHidden = true
        aktuelleDatei = name
        aktuellIstDatenbank = alsDatenbank
        if willSichern
        {
            empfänger.Sichern(datei: aktuelleDatei, alsDatenbank: alsDatenbank)
        }
        else
        {
            if empfänger.Lesen(datei: aktuelleDatei, alsDatenbank: alsDatenbank)
            {
            	graph.setNeedsDisplay()
            }
            else
        	{
        		aktuelleDatei = ""
            }
        }
    }
    
    /**
	Entfernt den Dateilesen-Dialog
    */
    func DateidialogAbbrechen()
    {
		NormalansichtZeigen()
    }
    
    /**
    Zeigt den Abfragedialog
	- parameters:
		- atktion: die Aktion für den Ja-Fall
    */
    func AbfrageZeigen(aktion: @escaping () -> Void)
    {
    	abfrageAktion = aktion
        hilfe.isHidden = true
        graph.isHidden = true
        werkzeugleiste.isHidden = true
        dateidialog.isHidden = true
        abfrage.isHidden = false
        knotenEditieren.isHidden = true
        kantenEditieren.isHidden = true
    }

    /**
	Entfernt den Abfrage-Dialog und führt die Aktion aus
    */
    func AbfragedialogJa()
    {
    	NormalansichtZeigen()
        abfrageAktion()
    }

    /**
	Entfernt den Abfrage-Dialog
    */
    func AbfragedialogNein()
    {
    	NormalansichtZeigen()
    }

	/**
	Zeigt einen Editieren-Dialog
	
	Je nachdem, ob ein Knoten oder eine Kante ausgewählt ist, wird der entsprechende Dialog angezeigt
    */
    func EditierenAnzeigen()
    {
    	if let k = steuerung.SelektiertenKnotenGeben()
    	{
    		knotenEditieren.DatenSetzen(bezeichner: k.BezeichnerGeben(), x: k.XGeben(), y: k.YGeben())
            hilfe.isHidden = true
            graph.isHidden = true
            werkzeugleiste.isHidden = true
            dateidialog.isHidden = true
            abfrage.isHidden = true
            knotenEditieren.isHidden = false
            kantenEditieren.isHidden = true
        }
        else if let k = steuerung.SelektierteKanteGeben()
        {
    		kantenEditieren.DatenSetzen(gewicht: k.GewichtGeben(), gerichtet: k.IstGerichtet())
            hilfe.isHidden = true
            graph.isHidden = true
            werkzeugleiste.isHidden = true
            dateidialog.isHidden = true
            abfrage.isHidden = true
            knotenEditieren.isHidden = true
            kantenEditieren.isHidden = false
        }
    }
    
    /**
	Entfernt den Knoteneditier-Dialog
    */
    func KnotendialogOK(bezeichner: String, x: CGFloat, y: CGFloat)
    {
		NormalansichtZeigen()
		let k = steuerung.SelektiertenKnotenGeben()!
    	k.BezeichnerSetzen(bezeichner: bezeichner)
    	k.PositionSetzen(x: x, y: y)
    }
    
    /**
	Entfernt den Knoteneditier-Dialog
    */
    func KnotendialogAbbrechen()
    {
		NormalansichtZeigen()
    }
    
    /**
	Entfernt den Kanteneditier-Dialog
    */
    func KantendialogOK(gewicht: String, gerichtet: Bool)
    {
		NormalansichtZeigen()
		let k = steuerung.SelektierteKanteGeben()!
    	k.GewichtSetzen(g: gewicht)
    	k.GerichtetSetzen(gerichtet: gerichtet)
    }
    
    /**
	Entfernt den Kanteneditier-Dialog
    */
    func KantendialogAbbrechen()
    {
		NormalansichtZeigen()
    }

    /**
    Löscht das Szenarium
    */
    func ZurücksetzenAusführen()
    {
        func Rücksetzen()
        {
            aktuelleDatei = ""
            empfänger.ZurückSetzen()
            graph.setNeedsDisplay()
        }
        
        if !empfänger.IstGeändert()
        {
        	Rücksetzen()
        }
        else
        {
        	AbfrageZeigen(aktion: Rücksetzen)
        }
    }
    
    /**
    Löscht das aktuell ausgewählte Objekt
    */
    func Löschen()
    {
        empfänger.Löschen()
    }
    
    /**
    Fragt ab, ob die Datei gesichert werden sollte.
	- parameters:
		- immer: der Sicherndialog wird immer aufgerufen
    */
    func SichernAusführen(immer: Bool)
    {
    	if aktuelleDatei.isEmpty || immer
    	{
    		willSichern = true
            graph.isHidden = true
            werkzeugleiste.isHidden = true
            hilfe.isHidden = true
            dateidialog.ÜberschriftSetzen(sichern: willSichern)
            dateidialog.isHidden = false
            abfrage.isHidden = true
            knotenEditieren.isHidden = true
            kantenEditieren.isHidden = true
        }
        else
        {
        	empfänger.Sichern(datei: aktuelleDatei, alsDatenbank: aktuellIstDatenbank)
        }
    }
    
    /**
    Öffnet eine Datenbank und liest den Inhalt in das Fenster.
    */
    func LesenAusführen ()
    {
    	func Lesen()
    	{
            willSichern = false
            graph.isHidden = true
            werkzeugleiste.isHidden = true
            hilfe.isHidden = true
            dateidialog.ÜberschriftSetzen(sichern: willSichern)
            dateidialog.isHidden = false
            abfrage.isHidden = true
            knotenEditieren.isHidden = true
            kantenEditieren.isHidden = true
        }
        
        if !empfänger.IstGeändert()
        {
        	Lesen()
        }
        else
        {
        	AbfrageZeigen(aktion: Lesen)
        }
    }
}

PlaygroundPage.current.liveView = OberflächenController()
